@@ -5,7 +5,6 @@ import (
 	"fmt"
 	"io"
 	"regexp"
-	"slices"
 	"strconv"
 	"strings"
 	"time"
@@ -526,20 +525,16 @@ func unpackNodes(node *yaml.Node) []*yaml.Node {
 	return nodes
 }
 
-func nodeKeys(node *yaml.Node) (keys []string) {
+func hasKey(node *yaml.Node, key string) bool {
 	if node.Kind != yaml.MappingNode {
-		return keys
+		return false
 	}
-	for i, n := range node.Content {
-		if i%2 == 0 && n.Value != "" {
-			keys = append(keys, n.Value)
+	for i := 0; i < len(node.Content); i += 2 {
+		if node.Content[i].Value == key {
+			return true
 		}
 	}
-	return keys
-}
-
-func hasKey(node *yaml.Node, key string) bool {
-	return slices.Contains(nodeKeys(node), key)
+	return false
 }
 
 func hasValue(node *YamlNode) bool {
@@ -637,14 +632,8 @@ type yamlMap struct {
 
 func mappingNodes(node *yaml.Node) []yamlMap {
 	m := make([]yamlMap, 0, len(node.Content)/2)
-	var key *yaml.Node
-	for _, child := range node.Content {
-		if key != nil {
-			m = append(m, yamlMap{key: key, val: child})
-			key = nil
-		} else {
-			key = child
-		}
+	for i := 0; i < len(node.Content); i += 2 {
+		m = append(m, yamlMap{key: node.Content[i], val: node.Content[i+1]})
 	}
 	return m
 }
