@@ -77,6 +77,38 @@ type DeadInfo struct {
 	Fragment posrange.PositionRange
 }
 
+type DeadLabelKind uint8
+
+const (
+	ImpossibleDeadLabel DeadLabelKind = iota
+	OrphanedLabel
+	DuplicatedJoin
+	UnusedLabel
+)
+
+func (dlk DeadLabelKind) String() string {
+	switch dlk {
+	case ImpossibleDeadLabel:
+		return "impossible label"
+	case OrphanedLabel:
+		return "orphaned label"
+	case DuplicatedJoin:
+		return "redundant label"
+	case UnusedLabel:
+		return "unused label"
+	}
+	return "unknown"
+}
+
+type DeadLabel struct {
+	Name          string
+	Reason        string
+	LabelReason   string
+	UsageFragment posrange.PositionRange
+	LabelFragment posrange.PositionRange
+	Kind          DeadLabelKind
+}
+
 type ReturnInfo struct {
 	LogicalExpr    string
 	ValuePosition  posrange.PositionRange
@@ -117,23 +149,24 @@ func MostOuterOperation[T promParser.Node](s Source) (T, bool) {
 }
 
 type Join struct {
-	On          []string
-	Ignoring    []string
-	AddedLabels []string
-	Src         Source              // The source we're joining with.
-	Op          promParser.ItemType // The binary operation used for this join.
-	Depth       int                 // Zero if this is a direct join, non-zero otherwise. sum(foo * bar) would be in-direct join.
+	MatchingLabels []string
+	AddedLabels    []string
+	Src            Source              // The source we're joining with.
+	Op             promParser.ItemType // The binary operation used for this join.
+	Depth          int                 // Zero if this is a direct join, non-zero otherwise. sum(foo * bar) would be in-direct join.
+	IsOn           bool
 }
 
 type Unless struct {
-	On       []string
-	Ignoring []string
-	Src      Source
+	MatchingLabels []string
+	Src            Source
+	IsOn           bool
 }
 
 type Source struct {
 	Labels        map[string]LabelTransform
 	DeadInfo      *DeadInfo
+	DeadLabels    []DeadLabel
 	Returns       promParser.ValueType
 	Operations    SourceOperations
 	Joins         []Join   // Any other sources this source joins with.
@@ -181,7 +214,7 @@ func (s Source) LabelExcludeReason(name string) (string, posrange.PositionRange)
 	return s.Labels[""].Reason, s.Labels[""].Fragment
 }
 
-func (s *Source) excludeAllLabels(reason string, fragment posrange.PositionRange, except []string) {
+func (s *Source) excludeAllLabels(expr, reason string, fragment, allFragment posrange.PositionRange, except []string) {
 	// Everything that was included until now but will be removed needs an explicit stamp to mark it as gone.
 	for name, l := range s.Labels {
 		if slices.Contains(except, name) {
@@ -206,7 +239,7 @@ func (s *Source) excludeAllLabels(reason string, fragment posrange.PositionRange
 			s.Labels[name] = LabelTransform{
 				Kind:     PossibleLabel,
 				Reason:   reason,
-				Fragment: fragment,
+				Fragment: FindArgumentPosition(expr, fragment, name),
 			}
 		} else {
 			r, f := s.LabelExcludeReason(name)
@@ -221,31 +254,59 @@ func (s *Source) excludeAllLabels(reason string, fragment posrange.PositionRange
 	s.Labels[""] = LabelTransform{
 		Kind:     ImpossibleLabel,
 		Reason:   reason,
-		Fragment: fragment,
+		Fragment: allFragment,
 	}
 	s.FixedLabels = true
 }
 
-func (s *Source) excludeLabel(reason string, fragment posrange.PositionRange, names ...string) {
+func (s *Source) excludeLabel(reason string, fragment posrange.PositionRange, name string) {
+	s.Labels[name] = LabelTransform{
+		Kind:     ImpossibleLabel,
+		Reason:   reason,
+		Fragment: fragment,
+	}
+}
+
+func (s *Source) joinLabels(expr string, within posrange.PositionRange, op promParser.ItemType, names []string, outside []posrange.PositionRange) {
 	for _, name := range names {
+		if l, ok := s.Labels[name]; ok && l.Kind == GuaranteedLabel {
+			s.DeadLabels = append(s.DeadLabels, DeadLabel{
+				Kind:   DuplicatedJoin,
+				Name:   name,
+				Reason: "Query is trying to join the `" + name + "` label that is already present on the other side of the query.",
+				UsageFragment: FindArgumentPosition(
+					expr,
+					FindFuncPosition(expr, within, promParser.ItemTypeStr[op], outside),
+					name,
+				),
+				LabelReason:   l.Reason,
+				LabelFragment: l.Fragment,
+			})
+			return
+		}
 		s.Labels[name] = LabelTransform{
-			Kind:     ImpossibleLabel,
-			Reason:   reason,
-			Fragment: fragment,
+			Kind: PossibleLabel,
+			Reason: fmt.Sprintf(
+				"Query is using `%s(%s)`, all labels included inside `%s(...)` will be joined to the results on the other side of the query.",
+				promParser.ItemTypeStr[op], strings.Join(names, ", "), promParser.ItemTypeStr[op],
+			),
+			Fragment: FindArgumentPosition(
+				expr,
+				FindFuncPosition(expr, within, promParser.ItemTypeStr[op], outside),
+				name,
+			),
 		}
 	}
 }
 
-func (s *Source) includeLabel(reason string, fragment posrange.PositionRange, names ...string) {
-	for _, name := range names {
-		if l, ok := s.Labels[name]; ok && l.Kind == GuaranteedLabel {
-			continue
-		}
-		s.Labels[name] = LabelTransform{
-			Kind:     PossibleLabel,
-			Reason:   reason,
-			Fragment: fragment,
-		}
+func (s *Source) includeLabel(expr, reason string, fragment posrange.PositionRange, name string) {
+	if l, ok := s.Labels[name]; ok && l.Kind == GuaranteedLabel {
+		return
+	}
+	s.Labels[name] = LabelTransform{
+		Kind:     PossibleLabel,
+		Reason:   reason,
+		Fragment: FindArgumentPosition(expr, fragment, name),
 	}
 }
 
@@ -259,6 +320,126 @@ func (s *Source) guaranteeLabel(reason string, fragment posrange.PositionRange,
 	}
 }
 
+func (s *Source) checkIncludedLabels(expr string, pos posrange.PositionRange, names []string) {
+	for _, name := range names {
+		if !s.CanHaveLabel(name) {
+			reason, fragment := s.LabelExcludeReason(name)
+			s.DeadLabels = append(s.DeadLabels, DeadLabel{
+				Kind:          ImpossibleDeadLabel,
+				Name:          name,
+				Reason:        "You can't use `" + name + "` because this label is not possible here.",
+				UsageFragment: FindArgumentPosition(expr, pos, name),
+				LabelReason:   reason,
+				LabelFragment: fragment,
+			})
+		}
+	}
+}
+
+func (s *Source) hasJoinUsingLabel(name string) bool {
+	for _, j := range s.Joins {
+		if j.IsOn && slices.Contains(j.MatchingLabels, name) {
+			return true
+		}
+		if !j.IsOn && !slices.Contains(j.MatchingLabels, name) {
+			return true
+		}
+	}
+	return false
+}
+
+func (s *Source) checkAggregationLabels(expr string, n *promParser.AggregateExpr) {
+	var pos posrange.PositionRange
+	switch {
+	case len(n.Grouping) == 0:
+		pos = FindFuncNamePosition(expr, n.PosRange, promParser.ItemTypeStr[n.Op])
+	case n.Without:
+		pos = FindFuncPosition(expr, n.PosRange, promParser.ItemTypeStr[promParser.WITHOUT], nil)
+	default:
+		pos = FindFuncPosition(expr, n.PosRange, promParser.ItemTypeStr[promParser.BY], nil)
+	}
+
+	for _, j := range s.Joins {
+		for _, name := range j.AddedLabels {
+			if s.hasJoinUsingLabel(name) {
+				// This label is used for some other join for our source.
+				continue
+			}
+			if !n.Without && slices.Contains(n.Grouping, name) {
+				continue
+			}
+			if n.Without && !slices.Contains(n.Grouping, name) {
+				continue
+			}
+			var (
+				labelPos    = pos
+				labelReason string
+			)
+			if t := s.findLabelTransform(name); t != nil {
+				labelPos = t.Fragment
+				labelReason = t.Reason
+			}
+			s.DeadLabels = append(s.DeadLabels, DeadLabel{
+				Kind:          UnusedLabel,
+				Name:          name,
+				Reason:        fmt.Sprintf("Previously joined label `%s` is being removed from the results.", name),
+				UsageFragment: FindArgumentPosition(expr, pos, name),
+				LabelReason:   labelReason,
+				LabelFragment: labelPos,
+			})
+		}
+	}
+}
+
+func (s *Source) findLabelTransform(name string) *LabelTransform {
+	if t, ok := s.Labels[name]; ok {
+		return &t
+	}
+	for _, j := range s.Joins {
+		if t := j.Src.findLabelTransform(name); t != nil {
+			return t
+		}
+	}
+	return nil
+}
+
+func (s *Source) checkJoinedLabels(expr string, n *promParser.BinaryExpr, dst Source) (dead []DeadLabel) {
+	pos := findBinOpsOperatorPosition(expr, n, promParser.ItemTypeStr[n.Op])
+	for _, j := range s.Joins {
+		for _, name := range j.AddedLabels {
+			if slices.Contains(n.VectorMatching.Include, name) {
+				// label is included in group_left or group_right
+				continue
+			}
+			if n.VectorMatching.On && slices.Contains(n.VectorMatching.MatchingLabels, name) {
+				// label is included inside on(...)
+				continue
+			}
+			if !n.VectorMatching.On && !slices.Contains(n.VectorMatching.MatchingLabels, name) {
+				// label is NOT included inside ignoring(...)
+				continue
+			}
+			var (
+				labelPos    = pos
+				labelReason string
+			)
+			if t := dst.findLabelTransform(name); t != nil {
+				labelPos = t.Fragment
+				labelReason = t.Reason
+			}
+			dead = append(dead, DeadLabel{
+				Kind:          OrphanedLabel,
+				Name:          name,
+				Reason:        fmt.Sprintf("This binary operation prevents previously joined label `%s` from being added to the results.", name),
+				UsageFragment: pos,
+				LabelReason:   labelReason,
+				LabelFragment: labelPos,
+			})
+		}
+	}
+	return dead
+}
+
 type Visitor func(s Source, j *Join, u *Unless)
 
 func innerWalk(fn Visitor, s Source, j *Join, u *Unless) {
@@ -318,7 +499,7 @@ func walkNode(expr string, node promParser.Node) (src []Source) {
 		s.ReturnInfo.ReturnedNumber = n.Val
 		s.ReturnInfo.AlwaysReturns = true
 		s.ReturnInfo.ValuePosition = n.PosRange
-		s.excludeAllLabels("This query returns a number value with no labels.", n.PosRange, nil)
+		s.excludeAllLabels(expr, "This query returns a number value with no labels.", n.PosRange, n.PosRange, nil)
 		s.Position = n.PosRange
 		src = append(src, s)
 
@@ -331,7 +512,7 @@ func walkNode(expr string, node promParser.Node) (src []Source) {
 		s.Type = StringSource
 		s.Returns = promParser.ValueTypeString
 		s.ReturnInfo.AlwaysReturns = true
-		s.excludeAllLabels("This query returns a string value with no labels.", n.PosRange, nil)
+		s.excludeAllLabels(expr, "This query returns a string value with no labels.", n.PosRange, n.PosRange, nil)
 		s.Position = n.PosRange
 		src = append(src, s)
 
@@ -423,106 +604,10 @@ func walkAggregation(expr string, n *promParser.AggregateExpr) (src []Source) {
 	}
 
 	switch n.Op {
-	case promParser.SUM:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "sum",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
-	case promParser.MIN:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "min",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
-	case promParser.MAX:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "max",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
-	case promParser.AVG:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "avg",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
-	case promParser.GROUP:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "group",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
-	case promParser.STDDEV:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "stddev",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
-	case promParser.STDVAR:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "stdvar",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
-	case promParser.COUNT:
-		for _, s = range parseAggregation(expr, n) {
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "count",
-				Node:      n,
-				Arguments: args,
-			})
-			if n.Without || !slices.Contains(n.Grouping, labels.MetricName) {
-				s.excludeLabel("Aggregation removes metric name.", n.PosRange, labels.MetricName)
-			}
-			src = append(src, s)
-		}
 	case promParser.COUNT_VALUES:
 		for _, s = range parseAggregation(expr, n) {
 			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "count_values",
+				Operation: promParser.ItemTypeStr[n.Op],
 				Node:      n,
 				Arguments: args,
 			})
@@ -537,10 +622,23 @@ func walkAggregation(expr string, n *promParser.AggregateExpr) (src []Source) {
 			}
 			src = append(src, s)
 		}
-	case promParser.QUANTILE:
+	case promParser.TOPK, promParser.BOTTOMK:
+		for _, s = range walkNode(expr, n.Expr) {
+			for i := range s.Joins {
+				s.Joins[i].Depth++
+			}
+			s.Type = AggregateSource
+			s.Operations = append(s.Operations, SourceOperation{
+				Operation: promParser.ItemTypeStr[n.Op],
+				Node:      n,
+				Arguments: args,
+			})
+			src = append(src, s)
+		}
+	default:
 		for _, s = range parseAggregation(expr, n) {
 			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "quantile",
+				Operation: promParser.ItemTypeStr[n.Op],
 				Node:      n,
 				Arguments: args,
 			})
@@ -549,32 +647,6 @@ func walkAggregation(expr string, n *promParser.AggregateExpr) (src []Source) {
 			}
 			src = append(src, s)
 		}
-	case promParser.TOPK:
-		for _, s = range walkNode(expr, n.Expr) {
-			for i := range s.Joins {
-				s.Joins[i].Depth++
-			}
-			s.Type = AggregateSource
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "topk",
-				Node:      n,
-				Arguments: args,
-			})
-			src = append(src, s)
-		}
-	case promParser.BOTTOMK:
-		for _, s = range walkNode(expr, n.Expr) {
-			for i := range s.Joins {
-				s.Joins[i].Depth++
-			}
-			s.Type = AggregateSource
-			s.Operations = append(s.Operations, SourceOperation{
-				Operation: "bottomk",
-				Node:      n,
-				Arguments: args,
-			})
-			src = append(src, s)
-		}
 		/*
 			TODO these are experimental and promParser.EnableExperimentalFunctions must be set to true to enable parsing of these.
 				case promParser.LIMITK:
@@ -604,25 +676,42 @@ func parseAggregation(expr string, n *promParser.AggregateExpr) (src []Source) {
 			s.Joins[i].Depth++
 		}
 
+		s.checkAggregationLabels(expr, n)
 		if n.Without {
-			s.excludeLabel(
-				fmt.Sprintf("Query is using aggregation with `without(%s)`, all labels included inside `without(...)` will be removed from the results.",
-					strings.Join(n.Grouping, ", ")),
-				FindPosition(expr, n.PosRange, "without"),
-				n.Grouping...,
-			)
+			for _, name := range n.Grouping {
+				s.excludeLabel(
+					fmt.Sprintf("Query is using aggregation with `%s(%s)`, all labels included inside `%s(...)` will be removed from the results.",
+						promParser.ItemTypeStr[promParser.WITHOUT], strings.Join(n.Grouping, ", "), promParser.ItemTypeStr[promParser.WITHOUT]),
+					FindArgumentPosition(
+						expr,
+						FindFuncPosition(expr, n.PosRange, promParser.ItemTypeStr[promParser.WITHOUT], nil),
+						name,
+					),
+					name,
+				)
+			}
 		} else {
 			if len(n.Grouping) == 0 {
+				funcNamePos := FindFuncNamePosition(expr, n.PosRange, promParser.ItemTypeStr[n.Op])
 				s.excludeAllLabels(
+					expr,
 					"Query is using aggregation that removes all labels.",
-					FindPosition(expr, n.PosRange, "sum"),
+					funcNamePos,
+					funcNamePos,
 					nil,
 				)
 			} else {
+				s.checkIncludedLabels(
+					expr,
+					FindFuncPosition(expr, n.PosRange, promParser.ItemTypeStr[promParser.BY], nil),
+					n.Grouping,
+				)
 				s.excludeAllLabels(
-					fmt.Sprintf("Query is using aggregation with `by(%s)`, only labels included inside `by(...)` will be present on the results.",
-						strings.Join(n.Grouping, ", ")),
-					FindPosition(expr, n.PosRange, "by"),
+					expr,
+					fmt.Sprintf("Query is using aggregation with `%s(%s)`, only labels included inside `%s(...)` will be present on the results.",
+						promParser.ItemTypeStr[promParser.BY], strings.Join(n.Grouping, ", "), promParser.ItemTypeStr[promParser.BY]),
+					FindFuncPosition(expr, n.PosRange, promParser.ItemTypeStr[promParser.BY], nil),
+					FindFuncNamePosition(expr, n.PosRange, promParser.ItemTypeStr[promParser.BY]),
 					n.Grouping,
 				)
 			}
@@ -680,14 +769,17 @@ func parsePromQLFunc(s Source, expr string, n *promParser.Call) Source {
 		s.Returns = promParser.ValueTypeVector
 		vs, _ := MostOuterOperation[*promParser.VectorSelector](s)
 		names := labelsFromSelectors([]labels.MatchType{labels.MatchEqual}, vs)
+		funcNamePos := FindFuncNamePosition(expr, n.PosRange, n.Func.Name)
 		s.excludeAllLabels(
+			expr,
 			fmt.Sprintf(`The [%s()](https://prometheus.io/docs/prometheus/latest/querying/functions/#%s) function is used to check if provided query doesn't match any time series.
 You will only get any results back if the metric selector you pass doesn't match anything.
 Since there are no matching time series there are also no labels. If some time series is missing you cannot read its labels.
 This means that the only labels you can get back from absent call are the ones you pass to it.
 If you're hoping to get instance specific labels this way and alert when some target is down then that won't work, use the `+"`up`"+` metric instead.`,
 				n.Func.Name, n.Func.Name),
-			FindPosition(expr, n.PosRange, n.Func.Name),
+			funcNamePos,
+			funcNamePos,
 			names,
 		)
 		s.guaranteeLabel(
@@ -713,9 +805,11 @@ If you're hoping to get instance specific labels this way and alert when some ta
 		if len(n.Args) == 0 {
 			s.ReturnInfo.AlwaysReturns = true
 			s.excludeAllLabels(
+				expr,
 				fmt.Sprintf("Calling `%s()` with no arguments will return an empty time series with no labels.",
 					n.Func.Name),
 				n.PosRange,
+				n.PosRange,
 				nil,
 			)
 		} else {
@@ -781,17 +875,22 @@ If you're hoping to get instance specific labels this way and alert when some ta
 		s.ReturnInfo.ReturnedNumber = math.Pi
 		s.ReturnInfo.ValuePosition = n.PosRange
 		s.excludeAllLabels(
+			expr,
 			fmt.Sprintf("Calling `%s()` will return a scalar value with no labels.", n.Func.Name),
 			n.PosRange,
+			n.PosRange,
 			nil,
 		)
 
 	case "scalar":
 		s.Returns = promParser.ValueTypeScalar
 		s.ReturnInfo.AlwaysReturns = true
+		funcPos := FindFuncPosition(expr, n.PositionRange(), n.Func.Name, nil)
 		s.excludeAllLabels(
+			expr,
 			fmt.Sprintf("Calling `%s()` will return a scalar value with no labels.", n.Func.Name),
-			FindPosition(expr, n.PositionRange(), n.Func.Name),
+			funcPos,
+			funcPos,
 			nil,
 		)
 
@@ -804,8 +903,10 @@ If you're hoping to get instance specific labels this way and alert when some ta
 		s.ReturnInfo.AlwaysReturns = true
 		s.ReturnInfo.ValuePosition = n.PosRange
 		s.excludeAllLabels(
+			expr,
 			fmt.Sprintf("Calling `%s()` will return a scalar value with no labels.", n.Func.Name),
 			n.PosRange,
+			n.PosRange,
 			nil,
 		)
 
@@ -829,9 +930,12 @@ If you're hoping to get instance specific labels this way and alert when some ta
 				s.ReturnInfo.KnownReturn = true
 			}
 		}
+		funcNamePos := FindFuncNamePosition(expr, n.PosRange, n.Func.Name)
 		s.excludeAllLabels(
+			expr,
 			fmt.Sprintf("Calling `%s()` will return a vector value with no labels.", n.Func.Name),
-			FindPosition(expr, n.PosRange, n.Func.Name),
+			funcNamePos,
+			funcNamePos,
 			nil,
 		)
 
@@ -895,6 +999,7 @@ func parseCall(expr string, n *promParser.Call) (src []Source) {
 }
 
 func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
+	pos := n.PositionRange()
 	src = make([]Source, 0, 2)
 	switch {
 	// foo{} + 1
@@ -929,29 +1034,43 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 
 		// foo{} +               bar{}
 		// foo{} + on(...)       bar{}
-		// foo{} + ignoring(...) bar{}
+		// foo{} * ignoring(...) bar{}
 		// foo{} /               bar{}
 	case n.VectorMatching.Card == promParser.CardOneToOne:
 		rhs := walkNode(expr, n.RHS)
 		for _, ls := range walkNode(expr, n.LHS) {
 			if n.VectorMatching.On {
+				ls.checkIncludedLabels(expr, pos, n.VectorMatching.MatchingLabels)
+				funcPos := FindFuncPosition(expr, pos, promParser.ItemTypeStr[promParser.ON], []posrange.PositionRange{
+					n.LHS.PositionRange(), n.RHS.PositionRange(),
+				})
 				ls.excludeAllLabels(
+					expr,
 					fmt.Sprintf(
-						"Query is using %s vector matching with `on(%s)`, only labels included inside `on(...)` will be present on the results.",
-						n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
+						"Query is using %s vector matching with `%s(%s)`, only labels included inside `%s(...)` will be present on the results.",
+						n.VectorMatching.Card, promParser.ItemTypeStr[promParser.ON], strings.Join(n.VectorMatching.MatchingLabels, ", "), promParser.ItemTypeStr[promParser.ON],
 					),
-					FindPosition(expr, n.PositionRange(), "on"),
+					funcPos,
+					funcPos,
 					n.VectorMatching.MatchingLabels,
 				)
 			} else {
-				ls.excludeLabel(
-					fmt.Sprintf(
-						"Query is using %s vector matching with `ignoring(%s)`, all labels included inside `ignoring(...)` will be removed on the results.",
-						n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
-					),
-					FindPosition(expr, n.PositionRange(), "ignoring"),
-					n.VectorMatching.MatchingLabels...,
-				)
+				for _, name := range n.VectorMatching.MatchingLabels {
+					ls.excludeLabel(
+						fmt.Sprintf(
+							"Query is using %s vector matching with `%s(%s)`, all labels included inside `%s(...)` will be removed on the results.",
+							n.VectorMatching.Card, promParser.ItemTypeStr[promParser.IGNORING], strings.Join(n.VectorMatching.MatchingLabels, ", "), promParser.ItemTypeStr[promParser.IGNORING],
+						),
+						FindArgumentPosition(
+							expr,
+							FindFuncPosition(expr, pos, promParser.ItemTypeStr[promParser.IGNORING], []posrange.PositionRange{
+								n.LHS.PositionRange(), n.RHS.PositionRange(),
+							}),
+							name,
+						),
+						name,
+					)
+				}
 				for _, rs := range rhs {
 					rs.IsConditional, rs.ReturnInfo.IsReturnBool = checkConditions(rs, n.Op, n.ReturnBool)
 					if ls.ReturnInfo.AlwaysReturns && rs.ReturnInfo.AlwaysReturns && ls.ReturnInfo.KnownReturn && rs.ReturnInfo.KnownReturn {
@@ -961,21 +1080,24 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 				}
 			}
 			for _, rs := range rhs {
-				if ok, s, pos := canJoin(ls, rs, n.VectorMatching); !ok {
+				rs.DeadLabels = append(rs.DeadLabels, rs.checkJoinedLabels(expr, n, rs)...)
+				if ok, s, p := canJoin(ls, rs, n.VectorMatching); !ok {
 					rs.DeadInfo = &DeadInfo{
 						Reason:   s,
-						Fragment: pos,
+						Fragment: p,
 					}
 				}
 				ls.Joins = append(ls.Joins, Join{
-					Src:         rs,
-					Op:          n.Op,
-					Depth:       0,
-					On:          onLabels(n.VectorMatching),
-					Ignoring:    ignoringLabels(n.VectorMatching),
-					AddedLabels: n.VectorMatching.Include,
+					Src:            rs,
+					Op:             n.Op,
+					Depth:          0,
+					MatchingLabels: n.VectorMatching.MatchingLabels,
+					AddedLabels:    nil,
+					IsOn:           n.VectorMatching.On,
 				})
 			}
+			ls.DeadLabels = append(ls.DeadLabels, ls.checkJoinedLabels(expr, n, ls)...)
+			ls.excludeLabel("Binary operation between two vectors removes metric names.", pos, labels.MetricName)
 			ls.IsConditional, ls.ReturnInfo.IsReturnBool = checkConditions(ls, n.Op, n.ReturnBool)
 			src = append(src, ls)
 		}
@@ -985,43 +1107,58 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 	case n.VectorMatching.Card == promParser.CardOneToMany:
 		lhs := walkNode(expr, n.LHS)
 		for _, rs := range walkNode(expr, n.RHS) {
-			rs.includeLabel(
-				fmt.Sprintf(
-					"Query is using %s vector matching with `group_right(%s)`, all labels included inside `group_right(...)` will be include on the results.",
-					n.VectorMatching.Card, strings.Join(n.VectorMatching.Include, ", "),
-				),
-				FindPosition(expr, n.PositionRange(), "group_right"),
-				n.VectorMatching.Include...,
-			)
+			rs.joinLabels(expr, pos, promParser.GROUP_RIGHT, n.VectorMatching.Include, []posrange.PositionRange{
+				n.LHS.PositionRange(), n.RHS.PositionRange(),
+			})
 			// If we have:
 			// foo * on(instance) group_left(a,b) bar{x="y"}
 			// then only group_left() labels will be included.
 			if n.VectorMatching.On {
-				rs.includeLabel(
-					fmt.Sprintf(
-						"Query is using %s vector matching with `on(%s)`, labels included inside `on(...)` will be present on the results.",
-						n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
-					),
-					FindPosition(expr, n.PositionRange(), "on"),
-					n.VectorMatching.MatchingLabels...,
-				)
+				rs.checkIncludedLabels(expr, pos, n.VectorMatching.MatchingLabels)
+				for _, name := range n.VectorMatching.MatchingLabels {
+					rs.includeLabel(
+						expr,
+						fmt.Sprintf(
+							"Query is using %s vector matching with `on(%s)`, labels included inside `on(...)` will be present on the results.",
+							n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
+						),
+						FindArgumentPosition(
+							expr,
+							FindFuncPosition(expr, pos, promParser.ItemTypeStr[promParser.ON], []posrange.PositionRange{
+								n.LHS.PositionRange(), n.RHS.PositionRange(),
+							}),
+							name,
+						),
+						name,
+					)
+				}
 			}
 			for _, ls := range lhs {
-				if ok, s, pos := canJoin(rs, ls, n.VectorMatching); !ok {
+				ls.checkIncludedLabels(
+					expr,
+					FindFuncPosition(expr, pos, promParser.ItemTypeStr[promParser.GROUP_RIGHT], []posrange.PositionRange{
+						n.LHS.PositionRange(),
+						n.RHS.PositionRange(),
+					}),
+					n.VectorMatching.Include,
+				)
+				rs.DeadLabels = append(rs.DeadLabels, ls.checkJoinedLabels(expr, n, rs)...)
+				if ok, s, p := canJoin(rs, ls, n.VectorMatching); !ok {
 					ls.DeadInfo = &DeadInfo{
 						Reason:   s,
-						Fragment: pos,
+						Fragment: p,
 					}
 				}
 				rs.Joins = append(rs.Joins, Join{
-					Src:         ls,
-					Op:          n.Op,
-					Depth:       0,
-					On:          onLabels(n.VectorMatching),
-					Ignoring:    ignoringLabels(n.VectorMatching),
-					AddedLabels: n.VectorMatching.Include,
+					Src:            ls,
+					Op:             n.Op,
+					Depth:          0,
+					MatchingLabels: n.VectorMatching.MatchingLabels,
+					AddedLabels:    n.VectorMatching.Include,
+					IsOn:           n.VectorMatching.On,
 				})
 			}
+			rs.excludeLabel("Binary operation between two vectors removes metric names.", pos, labels.MetricName)
 			rs.IsConditional, rs.ReturnInfo.IsReturnBool = checkConditions(rs, n.Op, n.ReturnBool)
 			src = append(src, rs)
 		}
@@ -1031,46 +1168,62 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 	case n.VectorMatching.Card == promParser.CardManyToOne:
 		rhs := walkNode(expr, n.RHS)
 		for _, ls := range walkNode(expr, n.LHS) {
-			ls.includeLabel(
-				fmt.Sprintf(
-					"Query is using %s vector matching with `group_left(%s)`, all labels included inside `group_left(...)` will be include on the results.",
-					n.VectorMatching.Card, strings.Join(n.VectorMatching.Include, ", "),
-				),
-				FindPosition(expr, n.PositionRange(), "group_left"),
-				n.VectorMatching.Include...,
-			)
+			ls.joinLabels(expr, pos, promParser.GROUP_LEFT, n.VectorMatching.Include, []posrange.PositionRange{
+				n.LHS.PositionRange(), n.RHS.PositionRange(),
+			})
 			if n.VectorMatching.On {
-				ls.includeLabel(
-					fmt.Sprintf(
-						"Query is using %s vector matching with `on(%s)`, labels included inside `on(...)` will be present on the results.",
-						n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
-					),
-					FindPosition(expr, n.PositionRange(), "on"),
-					n.VectorMatching.MatchingLabels...,
-				)
+				ls.checkIncludedLabels(expr, pos, n.VectorMatching.MatchingLabels)
+				for _, name := range n.VectorMatching.MatchingLabels {
+					ls.includeLabel(
+						expr,
+						fmt.Sprintf(
+							"Query is using %s vector matching with `on(%s)`, labels included inside `on(...)` will be present on the results.",
+							n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
+						),
+						FindArgumentPosition(
+							expr,
+							FindFuncPosition(expr, pos, promParser.ItemTypeStr[promParser.ON], []posrange.PositionRange{
+								n.LHS.PositionRange(), n.RHS.PositionRange(),
+							}),
+							name,
+						),
+						name,
+					)
+				}
 			}
 			for _, rs := range rhs {
-				if ok, s, pos := canJoin(ls, rs, n.VectorMatching); !ok {
+				rs.checkIncludedLabels(
+					expr,
+					FindFuncPosition(expr, pos, promParser.ItemTypeStr[promParser.GROUP_LEFT], []posrange.PositionRange{
+						n.LHS.PositionRange(),
+						n.RHS.PositionRange(),
+					}),
+					n.VectorMatching.Include,
+				)
+				ls.DeadLabels = append(ls.DeadLabels, rs.checkJoinedLabels(expr, n, ls)...)
+				if ok, s, p := canJoin(ls, rs, n.VectorMatching); !ok {
 					rs.DeadInfo = &DeadInfo{
 						Reason:   s,
-						Fragment: pos,
+						Fragment: p,
 					}
 				}
 				ls.Joins = append(ls.Joins, Join{
-					Src:         rs,
-					Op:          n.Op,
-					Depth:       0,
-					On:          onLabels(n.VectorMatching),
-					Ignoring:    ignoringLabels(n.VectorMatching),
-					AddedLabels: n.VectorMatching.Include,
+					Src:            rs,
+					Op:             n.Op,
+					Depth:          0,
+					MatchingLabels: n.VectorMatching.MatchingLabels,
+					AddedLabels:    n.VectorMatching.Include,
+					IsOn:           n.VectorMatching.On,
 				})
 			}
+			ls.excludeLabel("Binary operation between two vectors removes metric names.", pos, labels.MetricName)
 			ls.IsConditional, ls.ReturnInfo.IsReturnBool = checkConditions(ls, n.Op, n.ReturnBool)
 			src = append(src, ls)
 		}
 
 		// foo{} and on(...)       bar{}
 		// foo{} and ignoring(...) bar{}
+		// foo{} and bar{}
 		// foo{} unless bar{}
 	case n.VectorMatching.Card == promParser.CardManyToMany:
 		var lhsCanBeEmpty bool // true if any of the LHS query can produce empty results.
@@ -1078,14 +1231,24 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 		for _, ls := range walkNode(expr, n.LHS) {
 			var rhsConditional bool
 			if n.VectorMatching.On {
-				ls.includeLabel(
-					fmt.Sprintf(
-						"Query is using %s vector matching with `on(%s)`, labels included inside `on(...)` will be present on the results.",
-						n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
-					),
-					FindPosition(expr, n.PositionRange(), "on"),
-					n.VectorMatching.MatchingLabels...,
-				)
+				ls.checkIncludedLabels(expr, pos, n.VectorMatching.MatchingLabels)
+				for _, name := range n.VectorMatching.MatchingLabels {
+					ls.includeLabel(
+						expr,
+						fmt.Sprintf(
+							"Query is using %s vector matching with `on(%s)`, labels included inside `on(...)` will be present on the results.",
+							n.VectorMatching.Card, strings.Join(n.VectorMatching.MatchingLabels, ", "),
+						),
+						FindArgumentPosition(
+							expr,
+							FindFuncPosition(expr, pos, promParser.ItemTypeStr[promParser.ON], []posrange.PositionRange{
+								n.LHS.PositionRange(), n.RHS.PositionRange(),
+							}),
+							name,
+						),
+						name,
+					)
+				}
 			}
 			if !ls.ReturnInfo.AlwaysReturns || ls.IsConditional {
 				lhsCanBeEmpty = true
@@ -1095,10 +1258,10 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 				if isConditional {
 					rhsConditional = true
 				}
-				if ok, s, pos := canJoin(ls, rs, n.VectorMatching); !ok {
+				if ok, s, p := canJoin(ls, rs, n.VectorMatching); !ok {
 					rs.DeadInfo = &DeadInfo{
 						Reason:   s,
-						Fragment: pos,
+						Fragment: p,
 					}
 				}
 				switch {
@@ -1110,20 +1273,21 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 						}
 					}
 					ls.Unless = append(ls.Unless, Unless{
-						Src:      rs,
-						On:       onLabels(n.VectorMatching),
-						Ignoring: ignoringLabels(n.VectorMatching),
+						Src:            rs,
+						MatchingLabels: n.VectorMatching.MatchingLabels,
+						IsOn:           n.VectorMatching.On,
 					})
 				case n.Op != promParser.LOR:
 					ls.Joins = append(ls.Joins, Join{
-						Src:         rs,
-						Op:          n.Op,
-						Depth:       0,
-						On:          onLabels(n.VectorMatching),
-						Ignoring:    ignoringLabels(n.VectorMatching),
-						AddedLabels: n.VectorMatching.Include,
+						Src:            rs,
+						Op:             n.Op,
+						Depth:          0,
+						MatchingLabels: n.VectorMatching.MatchingLabels,
+						AddedLabels:    nil,
+						IsOn:           n.VectorMatching.On,
 					})
 				}
+				ls.DeadLabels = append(ls.DeadLabels, rs.checkJoinedLabels(expr, n, ls)...)
 			}
 			if n.Op == promParser.LAND && rhsConditional {
 				ls.IsConditional = true
@@ -1146,20 +1310,6 @@ func parseBinOps(expr string, n *promParser.BinaryExpr) (src []Source) {
 	return src
 }
 
-func onLabels(vm *promParser.VectorMatching) []string {
-	if vm.On {
-		return vm.MatchingLabels
-	}
-	return nil
-}
-
-func ignoringLabels(vm *promParser.VectorMatching) []string {
-	if !vm.On {
-		return vm.MatchingLabels
-	}
-	return nil
-}
-
 func checkConditions(s Source, op promParser.ItemType, isBool bool) (isConditional, isReturnBool bool) {
 	if !s.IsConditional && isBool {
 		isReturnBool = isBool
@@ -1304,9 +1454,17 @@ func formatDesc(expr string, ls, rs Source, op string) string {
 	return lse + " " + op + " " + rse
 }
 
-// FIXME sum() on ().
-func FindPosition(expr string, within posrange.PositionRange, fn string) posrange.PositionRange {
-	re := regexp.MustCompile("(?i)(" + regexp.QuoteMeta(fn) + ")[ \n\t]*\\(")
+func isOutside(pos posrange.PositionRange, outside []posrange.PositionRange) bool {
+	for _, out := range outside {
+		if pos.Start >= out.Start && pos.End <= out.End {
+			return false
+		}
+	}
+	return true
+}
+
+func FindFuncNamePosition(expr string, within posrange.PositionRange, fn string) posrange.PositionRange {
+	re := regexp.MustCompile("(?si)(" + regexp.QuoteMeta(fn) + ")(?:[ \n\t]*?)\\(")
 	idx := re.FindStringSubmatchIndex(GetQueryFragment(expr, within))
 	if idx == nil {
 		return within
@@ -1316,3 +1474,49 @@ func FindPosition(expr string, within posrange.PositionRange, fn string) posrang
 		End:   within.Start + posrange.Pos(idx[1]-1),
 	}
 }
+
+func FindFuncPosition(expr string, within posrange.PositionRange, fn string, outside []posrange.PositionRange) posrange.PositionRange {
+	re := regexp.MustCompile("(?si)(" + regexp.QuoteMeta(fn) + ")(?:[ \n\t]*?)\\((?:.*?)\\)")
+	idx := re.FindStringSubmatchIndex(GetQueryFragment(expr, within))
+	if idx == nil {
+		return within
+	}
+	var pos posrange.PositionRange
+	for chk := range slices.Chunk(idx, 2) {
+		pos = posrange.PositionRange{
+			Start: within.Start + posrange.Pos(chk[0]),
+			End:   within.Start + posrange.Pos(chk[1]),
+		}
+		if isOutside(pos, outside) {
+			return pos
+		}
+	}
+	return within
+}
+
+func FindArgumentPosition(expr string, within posrange.PositionRange, name string) posrange.PositionRange {
+	re := regexp.MustCompile("(?s)\\((?:(.*,?))(?:[ \n\t]*?)(" + regexp.QuoteMeta(name) + ")(?:[ \n\t]*?)(?:(,.*)?)\\)")
+	idx := re.FindStringSubmatchIndex(GetQueryFragment(expr, within))
+	if idx == nil {
+		return within
+	}
+	return posrange.PositionRange{
+		Start: within.Start + posrange.Pos(idx[4]),
+		End:   within.Start + posrange.Pos(idx[5]),
+	}
+}
+
+func findBinOpsOperatorPosition(expr string, n *promParser.BinaryExpr, op string) posrange.PositionRange {
+	within := posrange.PositionRange{
+		Start: n.LHS.PositionRange().End + 1,
+		End:   n.RHS.PositionRange().Start,
+	}
+	idx := strings.Index(GetQueryFragment(expr, within), op)
+	if idx < 0 {
+		return within
+	}
+	return posrange.PositionRange{
+		Start: within.Start + posrange.Pos(idx),
+		End:   within.Start + posrange.Pos(idx+len(op)),
+	}
+}
